// Metrics: 6 -- Pixels
/*
   SQLFormControls.cc
	Author....: Ken Mayer
	            Based on work by other developers, both from earlier
					versions of the dUFLP, and my own work, as well as
					discussions in the dBASE Newsgroups ... Some of these
					controls are based on :dUFLP:FormCntl.cc, including
					the constant values defined at the beginning of this file.
					
					These controls are meant to work with the Database
					classes on forms, but specifically to work with
					SQL Databases, including the ADO database classes. 
					
					While still attempting to get a grip on concepts 
					(and arguing with one developer who is didactic
					about his way of looking at databases), these are designed
					to create their own versions of buffers, and allow
					some control not contained in the original classes in the
					dUFLP .cc file: FormCntl.cc.
					
					This code was created for:
					   The dBASE Book Plus
					   2015, Ken Mayer
					
					Permission is granted for use both as a learning tool, and 
					for a developers' applications, as long as credit is given ...

   NOTE: These controls assume dBASE Plus 8 or later!!
*/

/*
   ****************************************************************
	DEFINE CONSTANTS FOR USE IN ALL CONTROLS ...
	Copied from :dUFLP:FormCntl.cc
	****************************************************************
*/

// ----------------------------------------------------
// Font info:
#DEFINE CONTROLFONTNAME "WindowText"    // default
#DEFINE S_CONTROLFONTSIZE 10            // For use with 'small fonts'
#DEFINE L_CONTROLFONTSIZE  8            // For use with 'large fonts'
#DEFINE S_LABELFONTSIZE   10            // For use with 'small fonts'
#DEFINE L_LABELFONTSIZE    8            // For use with 'large fonts'

// ----------------------------------------------------
// Colors:
/*
    Color Notes:
       To use Windows Schemes, so that your application matches
       the scheme your user might be using, you need to use
       information from the online help. This can be a bit
       deceptive, but ...

       By default forms use:             BtnFace (color of buttons)
       The standard Text color is:       BtnText
       Standard Highlight color is:      Highlight
       Standard Highlight text color is: HighlightText
       Data Entry is normally:           WindowText/Window
           (Color of text in windows, background is standard
            window color ...)
       The following is from the online help:

                             Corresponding Display Properties 
       Color                 Appearance Color
       ========================================================
       ActiveBorder          Active window border
       ActiveCaption         Active title bar
       AppWorkspace          Application background
       Background            Desktop
       BtnFace               3D objects
       BtnHighlight          A shade lighter than 3D objects
       BtnShadow             A shade darker than 3D objects
       BtnText               3D objects font
       CaptionText           Active title bar font
       GrayText              preset gray -- not available
       Highlight             Selected items
       HighlightText         Selected items font
       InactiveBorder        Inactive window border
       InactiveCaption       Inactive title bar
       InactiveCaptionText   Inactive title bar font
       InfoText              ToolTip
       InfoBk                ToolTip font
       Menu                  Menu
       MenuText              Menu font
       Scrollbar             A shade lighter than 3D objects
       Window                Window
       WindowFrame           preset drop shadow -- not available
       WindowText            Window font

       Some basic testing of the settings given below with 
       the code onDesignOpen for many of the controls, 
       appears to set the colors properly for when a user 
       wants to use Windows Schemes other than basic 
       default ones, except as noted (highlight is different, 
       used for things like colorHighlight for controls, 
       rowSelectColor for the grid ...). 
*/

// For most controls:
#DEFINE HIGHLIGHT       "WindowText/0x80ffff" // yellow background
#DEFINE NORMAL          "WindowText/Window"
#DEFINE NORMALCHECK     "BtnText/BtnFace" // Checkbox and RadioButton Normal
// can be used for read-only as well ...:
#DEFINE DISABLED        "Gray/Window"
// new for dBASE Plus 8-10, and SQL/ADO ...
#DEFINE FIELDMODIFIED   "Blue/Window"    // blue on standard window color
#DEFINE FIELDMODHIGH    "Blue/0x80ffff" // blue on yellow when highlighted and changed
// for objects with errors
#DEFINE ERRORCOLOR      "WHITE/RED"

// ----------------------------------------------------
// Standard border for entry objects:
#DEFINE ENTRYBORDERSTYLE  7 // used in SQLCustEntryField, SQLCustEditor,
                            // SQLCustImage, SQLCustGrid, SQLCustCombobox
                            // SQLCustTreeView
#DEFINE CHECKBORDER       3 // SQLCustCheckBox, SQLCustRadio
      /* border options (so you don't have to go look 'em up):
             0 = default
             1 = raised
             2 = lowered
             3 = none
             4 = single (single-line)
             5 = double (double-line)
             6 = drop shadow
             7 = client
             8 = modal
             9 = etched in
             10 = etched out
       */
// ----------------------------------------------------
// end of definitions of constants ...
// ----------------------------------------------------

/*
   ----------------------------------------------------------------
   SQLTextLabel -- used with various controls, it is 
	simply a place to store the name of the field, or
	reference whatever you need. This is a pretty 
	basic control, the text is aligned to the right,
	font is bold ...
   ----------------------------------------------------------------
*/
class SQLTextLabel( oParent ) of TextLabel( oParent ) custom
   with (this)
        alignHorizontal := 2 // Right
        alignVertical   := 1 // Middle
		  fontBold        := true
        fontName        := CONTROLFONTNAME
        if smallFonts()
           fontSize     := S_LABELFONTSIZE
        else
           fontSize     := L_LABELFONTSIZE
        endif
        metric          := 6 // Pixels
        text            := "SQLTextLabel:"
	endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use form metric of 6 (Pixels) "+;
                 "for this control.", "Metric Issue" )
         return
      endif
   return // onDesignOpen event
	
endclass

/* 
   ----------------------------------------------------------------
   SQLEntryField -- Standard entryfield. The onGotFocus event 
   can be over-ridden easily enough ... it's just useful to 
   be sure that the cursor isn't somewhere you don't want it.
   (Also in this version, unlike the dUFLP FormCntl.cc
   kmEntryfield class, this uses a timer object for checking
	for modifications ... see below.)
   I generally prefer selectAll to be off so that I (or a user) 
   don't over-write the contents of the entryfield by accident.
	
	Checking for modification of the Value:
	In addition, the control uses a timer object to handle checking
	for modification of the value property. The timer is set up
	in the onOpen event handler, when the control gains focus 
	(OnGotFocus), the timer is made active, when the control 
	loses focus (onLostFocus) the timer is made inactive. 
	This has the advantage that is lost when using the onKey event
	handler, in that it doesn't matter what keystrokes were used
	to modify the contents of the value (Ctrl+X, Ctrl+V were 
	not being trapped with onKey, among others), every half
	second the timer fires the method isModified, and we deal with
	there. 
	
	It is therefore vital that if you override the onGotFocus or 
	onLostFocus event handlers, that you put in a call to the
	ones here: SQLEntryfield::onGotFocus() or SQLEntryfield::onLostFocus().
	
	   Required:
		this.tableName       // required so we can work with multiple
		                     // tables on a form ...
		this.fieldName       // we are not using the dataLink property,
		                     // so need to know what field to get
									// data from, and assign to value property
		this.includeInQuery  // do we want to save the value back to the table
		                     // if it was modified?
		this.defaultValue    // when appending data, it is useful to have
		                     // a defaultValue set, so that the control
									// can deal with things. For example, 
									// a character field might default to "",
									// a date field might default to {}, etc.
									
	In a form's open or readModal (overridden versions) methods,
	you would have code along these lines:
	
		form.CustNumEF.tableName      := "Customers"
		form.CustNumEF.fieldName      := "CustomerNumber"
		form.CustNumEF.readOnly       := true
		form.CustNumEF.includeInQuery := false
		
		form.BusNameEF.tableName      := "Customers"
		form.BusNameEF.fieldName      := "BusinessName"
      form.BusNameEF.defaultValue   := ""
		// not a custom property, but extremely useful for
		// character fields, to limit the length of what
		// the user enters to the length of the field
		// itself. Very disconcerting to enter something
		// longer, and having the software cut it off when
		// the row is saved ...:
      form.BusNameEF.maxLength      := 30		

	Modified to include the buffering functionality that appears
	to be required by ADO/SQL-Server Databases (see custom 
	properties ...). 
   ----------------------------------------------------------------
*/
class SQLEntryField( oParent ) of Entryfield( oParent ) custom

   with (this)
        borderStyle    := ENTRYBORDERSTYLE
        colorNormal    := NORMAL
        colorHighlight := HIGHLIGHT
        fontName       := CONTROLFONTNAME
        if smallFonts()
           fontSize    := S_CONTROLFONTSIZE
        else
           fontSize    := L_CONTROLFONTSIZE
        endif
        height         := 22
        metric         := 6  // Pixels
        selectAll      := false
        value          := "SQLCustEntryField"
   endwith
	
	// ------------------------------------------------------------------------
	// Custom Properties:
   // TableName required to track specific table,
	// particularly when multiple tables in form
	this.tableName      = null
   // FieldName is used to keep track of which field
	// in the table is being used, since we're not 
	// dataLinking ...
	this.fieldName      = null
	
   // These two are used for the buffering needed, the values
	// need to be exactly these names, so that they can be called
	// from code in another routine that loads the data from a row
	// to the controls, and also allows you to save it back out.
	this.oldValue       = null
	this.newValue       = null
	
	// default value for when appending:
	this.defaultValue   = null

   // include this control in query object references
	// can be used to load data in form's open
	// event handler ...
	// Should it be checked when saving?
	this.includeInQuery = true

   // is the content of the field changed from
	// the original?
	this.modified       = false

   // is the entryfield readOnly?
	this.readOnly       = false
	
	// is there an error? (Needs to be checked 
	// in query's canSave event handler ...)
	this.error          = false

   /*
	   This code will execute when the control is instantiated,
		before the form opens ...
	*/
	// add code here to add this item to form's
	// aSQLControls array ... (if array doesn't 
	// exist, create it, then add this ...)
	if this.includeInQuery
		if type( "form.aSQLControls" ) == "U"
			form.aSQLControls = new array()
		endif // type(...)
		// add control to array
		form.aSQLControls.add( this ) // object reference
	endif // this.includeInQuery

	// Control Event Handlers:
   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 6
         msgbox( "Please use form metric of 6 (Pixels) "+;
                 "for this control.", "Metric Issue" )
         return
      endif
   return // onDesignOpen event

   function onOpen
	   SetReadOnly( this )
      // set the "old" value to value currently in field
		// this is "datalinked" to:
		if this.fieldName # null
		   this.oldValue := this.value
	   endif // not empty ...
		
      // Suggested by Heinz Kesting:
	   this.ModTimer = new timer()
	   this.ModTimer.interval := 0.5 // half second
      this.ModTimer.onTimer  := class::isModified
		this.ModTimer.parent    = this // need reference to this object
   return // onOpen event

   function onGotFocus
       this.keyboard( "{Home}" )
	    // set timer:	 
		 this.ModTimer.enabled := true
   return // onGotFocus event
	
	function onLostFocus
	   // turn timer off:
	   this.modTimer.enabled := false
	return

   /*
	   Attempts to set this method up as an independent
		function get weirder and weirder, as determining
		which object is calling the code, and so on
		becomes impossible. So ... we won't. 
	*/
   function isModified
	   // called from an object that doesn't understand form references:
		oForm = this.parent.form
		oThis = this.parent

      // set custom property of form:
		if type( "oForm.modified" ) # "L"
		   oForm.modified = false
		endif

      // if it's read-only we don't want to 
		// do anything to the colors:
		if oThis.ReadOnly
		   return
		endif

      // set the "new" value to value currently in entryfield
		// control, so that ultimately we can save it out to
		// the field ...
		oThis.newValue := oThis.value
		// change the color of the control if the value
		// is different than the original:
		if oThis.oldValue # oThis.newValue
		   if oThis.error
				// change the color
				oThis.colorNormal    := ERRORCOLOR
				oThis.colorHighlight := ERRORCOLOR
			else
				// change the color
				oThis.colorNormal    := FIELDMODIFIED
				oThis.colorHighlight := FIELDMODHIGH
				// set modified property:
				oThis.modified       := true
				// set form's property:
				oForm.modified       := true
			endif // this.error
		else 
		   oThis.colorNormal    := NORMAL
			oThis.colorHighlight := HIGHLIGHT
			oThis.modified       := false
		endif // this.oldValue ...
		
   return // isModified method

endclass 
// End of class: SQLEntryfield


// functions required by code above:

function SetReadonly( oObject )
	// is this a read-only control?
	if oObject.readOnly
		// set the color:
		oObject.colorNormal    := DISABLED
		// change mousepointer
		oObject.mousePointer   := 12 // No		
		// disable the control:
		oObject.when           := {; return false }
		oObject.newValue       := oObject.value
	endif // this.readOnly
return


/*
   Misc Functions called from classes above:
*/

Function Smallfonts // from MiscAPI.prg
/*
   Programmer   : Romain Strieff [dBVIPS]
   Last Modified: 04/20/1999
   Rev. History.: 04/20/1999 -- Original
   Notes........: Returns True if Smallfonts are used
   .............: in Windows menus
   Written for..: Visual dBASE (32-bit versions)
   Usage .......: ? SmallFonts()
*/
   local dc,result
   extern CINT        GetDeviceCaps(cHandle, CINT) gdi32
   extern CINT        ReleaseDC(cHandle, cHandle) user32
   extern cHandle     GetDC(cHandle) user32
   #define LOGPIXELSX    88
   dc=getdc(0)
   result=GetDeviceCaps(dc,LOGPIXELSX)
   ReleaseDc(0,dc)
return result == 96
// eof: SmallFonts()

function Repaint
/*
   ----------------------------------------------------------------
   This is a routine by Jim Sare with minor
   modifications by Ken Mayer. It can be used
   to ensure that if any processing has occurred
   that the form gets repainted (which sometimes
   doesn't happen) ...
   ----------------------------------------------------------------
*/
   parameter oForm
   if type( "oForm" ) == "U" or empty( oForm )
      oForm = form
   endif
   if type( "UpdateWindow" ) # "FP"
      extern CLOGICAL UpdateWindow( CHANDLE ) USER32
   endif
RETURN UpdateWindow( oForm.hWnd )
// end of function: Repaint()


/*
  -----------------------------------------------------------------------
  End of File: SQLFormControls.cc
  -----------------------------------------------------------------------
*/